@@ -9,67 +9,68 @@ Racket-like multi-input ``foldl`` and ``foldr`` based on
 and (stream-scan) in SRFI-41.
   https://srfi.schemers.org/srfi-41/srfi-41.html
 
-``take`` and ``drop`` based on Haskell.
-
 ``flatten`` based on Danny Yoo's version:
   http://rightfootin.blogspot.fi/2006/09/more-on-python-flatten.html
 """
 
 __all__ = ["scanl", "scanr", "scanl1", "scanr1",
            "foldl", "foldr", "reducel", "reducer",
-           "flatmap", "mapr", "zipr", "uniqify", "uniq",
+           "mapr", "zipr", "map_longest", "mapr_longest", "zipr_longest",
+           "flatmap", "uniqify", "uniq",
            "take", "drop", "split_at", "unpack",
            "tail", "first", "second", "nth", "last",
            "flatten", "flatten1", "flatten_in",
            "iterate", "iterate1"]
 
-from itertools import tee, islice
+from operator import itemgetter
+from functools import partial
+from itertools import tee, islice, zip_longest, starmap, chain, filterfalse, groupby
 from collections import deque
-from inspect import isgenerator
 
 # require at least one iterable to make this work seamlessly with curry.
-def scanl(proc, init, iterable0, *iterables):
+def scanl(proc, init, iterable0, *iterables, longest=False, fillvalue=None):
     """Scan (accumulate), optionally with multiple input iterables.
 
-    Similar to ``itertools.accumulate``. If the inputs are generators, this is
+    Similar to ``itertools.accumulate``. If the inputs are iterators, this is
     essentially a lazy ``foldl`` that yields the intermediate result at each step.
     Hence, useful for partially folding infinite sequences.
 
-    At least one iterable (``iterable0``) is required. More are optional.
-
-    Terminates when the shortest input runs out.
-
     Initial value is mandatory; there is no sane default for the case with
     multiple inputs.
 
-    Returns a generator, which (roughly, in pseudocode)::
+    At least one iterable (``iterable0``) is required. More are optional.
+
+    By default, terminate when the shortest input runs out. To terminate on
+    longest input, use ``longest=True`` and optionally provide a ``fillvalue``.
+
+    Returns a generator, which (in pseudocode)::
 
         acc = init
-        for elts in zip(iterable0, *iterables):
-            yield proc(*elts, acc)  # if this was legal syntax
-    """
-    iterables = (iterable0,) + iterables
-    def heads(its):
-        hs = []
-        for it in its:
-            try:
-                h = next(it)
-            except StopIteration:  # shortest sequence ran out
-                return StopIteration
-            hs.append(h)
-        return tuple(hs)
-    iters = tuple(iter(x) for x in iterables)
-    acc = init
-    while True:
         yield acc
-        hs = heads(iters)
-        if hs is StopIteration:
-            break
-        acc = proc(*(hs + (acc,)))
+        for elts in zip(iterable0, *iterables):  # or zip_longest as appropriate
+            acc = proc(*elts, acc)  # if this was legal syntax
+            yield acc
 
-def scanr(proc, init, sequence0, *sequences):
+    Example - partial sums and products::
+
+        from operator import add, mul
+        psums = composer(tail, curry(scanl, add, 0))  # tail to drop the init value
+        pprods = composer(tail, curry(scanl, mul, 1))
+        data = range(1, 5)
+        assert tuple(psums(data)) == (1, 3, 6, 10)
+        assert tuple(pprods(data)) == (1, 2, 6, 24)
+    """
+    z = zip if not longest else partial(zip_longest, fillvalue=fillvalue)
+    acc = init
+    yield acc
+    for elts in z(iterable0, *iterables):
+        acc = proc(*(elts + (acc,)))
+        yield acc
+
+def scanr(proc, init, sequence0, *sequences, longest=False, fillvalue=None):
     """Like scanl, but scan from the right (walk each sequence backwards)."""
-    return scanl(proc, init, reversed(sequence0), *(reversed(s) for s in sequences))
+    return scanl(proc, init, reversed(sequence0), *(reversed(s) for s in sequences),
+                 longest=longest, fillvalue=fillvalue)
 
 def scanl1(proc, iterable, init=None):
     """scanl for a single iterable, with optional init.
@@ -77,6 +78,15 @@ def scanl1(proc, iterable, init=None):
     If ``init is None``, use the first element from the iterable.
 
     If the iterable is empty, return ``None``.
+
+    Example - partial sums and products::
+
+        from operator import add, mul
+        psums = curry(scanl1, add)
+        pprods = curry(scanl1, mul)
+        data = range(1, 5)
+        assert tuple(psums(data)) == (1, 3, 6, 10)
+        assert tuple(pprods(data)) == (1, 2, 6, 24)
     """
     it = iter(iterable)
     if not init:
@@ -94,25 +104,28 @@ def scanr1(proc, sequence, init=None):
     """
     return scanl1(proc, reversed(sequence), init)
 
-def foldl(proc, init, iterable0, *iterables):
+def foldl(proc, init, iterable0, *iterables, longest=False, fillvalue=None):
     """Racket-like foldl that supports multiple input iterables.
 
-    At least one iterable (``iterable0``) is required. More are optional.
-
-    Terminates when the shortest input runs out.
-
     Initial value is mandatory; there is no sane default for the case with
     multiple inputs.
 
+    At least one iterable (``iterable0``) is required. More are optional.
+
+    By default, terminate when the shortest input runs out. To terminate on
+    longest input, use ``longest=True`` and optionally provide a ``fillvalue``.
+
     Note order: ``proc(elt, acc)``, which is the opposite order of arguments
     compared to ``functools.reduce``. General case ``proc(e1, ..., en, acc)``.
     """
-    return last(scanl(proc, init, iterable0, *iterables))
+    return last(scanl(proc, init, iterable0, *iterables,
+                      longest=longest, fillvalue=fillvalue))
 
-def foldr(proc, init, sequence0, *sequences):
+def foldr(proc, init, sequence0, *sequences, longest=False, fillvalue=None):
     """Like foldl, but fold from the right (walk each sequence backwards)."""
     # Reverse, then left-fold gives us a linear process.
-    return foldl(proc, init, reversed(sequence0), *(reversed(s) for s in sequences))
+    return foldl(proc, init, reversed(sequence0), *(reversed(s) for s in sequences),
+                 longest=longest, fillvalue=fillvalue)
 
 def reducel(proc, iterable, init=None):
     """Foldl for a single iterable, with optional init.
@@ -130,6 +143,30 @@ def reducer(proc, sequence, init=None):
     """
     return reducel(proc, reversed(sequence), init)
 
+def mapr(func, *sequences):
+    """Like map, but walk each sequence from the right."""
+    return map(func, *(reversed(s) for s in sequences))
+
+def zipr(*sequences):
+    """Like zip, but walk each sequence from the right."""
+    return zip(*(reversed(s) for s in sequences))
+
+def map_longest(func, *iterables, fillvalue=None):
+    """Like map, but terminate on the longest input.
+
+    In the input to ``func``, missing elements (after end of shorter inputs)
+    are replaced by ``fillvalue``, which defaults to ``None``.
+    """
+    return starmap(func, zip_longest(*iterables, fillvalue=fillvalue))
+
+def mapr_longest(func, *sequences, fillvalue=None):
+    """Like map_longest, but walk each sequence from the right."""
+    return map_longest(func, *(reversed(s) for s in sequences), fillvalue=fillvalue)
+
+def zipr_longest(*sequences, fillvalue=None):
+    """Like itertools.zip_longest, but walk each sequence from the right."""
+    return zip_longest(*(reversed(s) for s in sequences), fillvalue=fillvalue)
+
 def flatmap(f, iterable0, *iterables):
     """Map, then concatenate results.
 
@@ -138,7 +175,7 @@ def flatmap(f, iterable0, *iterables):
     ``f`` should accept as many arguments as iterables given (each argument
     drawn from one of the iterables), and return an iterable.
 
-    Returns a generator that yields the flatmapped result.
+    Returns an iterator that yields the flatmapped result.
 
     Example::
 
@@ -161,21 +198,9 @@ def flatmap(f, iterable0, *iterables):
         assert tuple(flatmap(sum_and_diff, (10, 20, 30), (1, 2, 3))) == \\
                (11, 9, 22, 18, 33, 27)
     """
-#    def concat(elt, acc):
-#        return tuple(acc) + tuple(elt)
-#    return foldl(concat, (), map(f, *lsts))  # eager, bad
-    iterables = (iterable0,) + iterables
-    for xs in map(f, *iterables):
-        for x in xs:
-            yield x
-
-def mapr(func, *sequences):
-    """Like map, but walk each sequence from the right."""
-    return map(func, *(reversed(s) for s in sequences))
-
-def zipr(*sequences):
-    """Like zip, but walk each sequence from the right."""
-    return zip(*(reversed(s) for s in sequences))
+    yield from chain.from_iterable(map(f, iterable0, *iterables))
+#    for xs in map(f, iterable0, *iterables):
+#        yield from xs
 
 def uniqify(iterable, key=None):
     """Skip duplicates in iterable.
@@ -185,74 +210,66 @@ def uniqify(iterable, key=None):
 
     If ``key`` is provided, the return value of ``key(elt)`` is tested instead
     of ``elt`` itself to determine uniqueness.
+
+    This is ``unique_everseen`` from ``itertools`` recipes.
     """
-    key = key or (lambda x: x)
     it = iter(iterable)
     seen = set()
-    for e in it:
-        k = key(e)
-        if k not in seen:
-            seen.add(k)
+    seen_add = seen.add
+    if key is None:
+        for e in filterfalse(seen.__contains__, it):
+            seen_add(e)
             yield e
+    else:
+        for e in it:
+            k = key(e)
+            if k not in seen:
+                seen_add(k)
+                yield e
 
 def uniq(iterable, key=None):
     """Like uniqify, but for consecutive duplicates only.
 
     Named after the *nix utility.
-    """
-    key = key or (lambda x: x)
-    it = iter(iterable)
-    lasthash = object()  # essentially gensym
-    for e in it:
-        h = hash(key(e))
-        if h != lasthash:  # no guarantees on singleton-ness
-            lasthash = h
-            yield e
 
-def _makegen(iterable):  # ensure iterable is a generator
-    if isgenerator(iterable):
-        return iterable
-    def gen():
-        yield from iterable
-    return gen()
+    This is ``unique_justseen`` from ``itertools`` recipes.
+    """
+    # the outer map retrieves the item from the subiterator in (key, subiterator).
+    return map(next, map(itemgetter(1), groupby(iterable, key)))
 
 def take(n, iterable):
-    """Return a generator that yields the first n items of iterable, then stops.
+    """Return an iterator that yields the first n items of iterable, then stops.
 
     Stops earlier if ``iterable`` has fewer than ``n`` items.
 
-    This is essentially ``take`` from ``itertools`` recipes,
-    but returns a generator.
+    This is ``take`` from ``itertools`` recipes.
     """
     if n < 0:
         raise ValueError("expected n >= 0, got {}".format(n))
-    it = iter(iterable)
-    it = islice(it, n)
-    return _makegen(it)
+    return islice(iter(iterable), n)
 
 def drop(n, iterable):
     """Skip the first n elements of iterable, then yield the rest.
 
     If ``n`` is ``None``, consume the iterable until it runs out.
 
-    This is essentially ``consume`` from ``itertools`` recipes,
-    but returns a generator.
+    This is ``consume`` from ``itertools`` recipes.
     """
     if n < 0:
         raise ValueError("expected n >= 0, got {}".format(n))
-    elif n == 0:
-        return iterable
     it = iter(iterable)
     if n is None:
         deque(it, maxlen=0)
     else:
         next(islice(it, n, n), None)  # advance it to empty slice starting at n
-    return _makegen(it)
+    return it
 
 def split_at(n, iterable):
     """Split iterable at position n.
 
-    Returns a pair of generators ``(first_part, second_part)``.
+    Returns a pair of iterators ``(first_part, second_part)``.
+
+    Based on ``itertools.tee``, ``take`` and ``drop``.
 
     Examples::
 
@@ -269,30 +286,32 @@ def split_at(n, iterable):
     ia, ib = tee(iter(iterable))
     return take(n, ia), drop(n, ib)
 
-# FIXME: difficult to make unpack curryable.
-def unpack(iterable, n, k=None, fillvalue=None):
+def unpack(n, iterable, *, k=None, fillvalue=None):
     """From iterable, return the first n elements, and the kth tail.
 
     This is a lazy generalization of sequence unpacking that works also for
     infinite iterables.
 
-    The return value is a tuple containing the ``n`` first elements, and as its
-    last item, the tail of the iterable from item ``k`` onwards.
-
     Default ``k=None`` means ``k = n``, i.e. return the tail that begins
     right after the extracted items. Other values are occasionally useful,
     e.g. to peek into the tail, while not permanently extracting an item.
 
+    The return value is a tuple containing the ``n`` first elements, and as its
+    last item, an iterator representing the tail of the iterable from item ``k``
+    onwards.
+
     If there are fewer than ``n`` items in the iterable, the missing items
-    are returned as ``fillvalue``. The ``rest`` part is then a generator
-    that just raises ``StopIteration``.
+    are returned as ``fillvalue``. The tail is then a generator that just
+    raises ``StopIteration``.
 
     If ``k < n`` (tail overlaps with the extracted items), the tail
     is formed by calling ``itertools.tee`` at the appropriate point
-    during the extraction.
+    during the extraction. (Plan the client code accordingly; see the
+    caution in `itertools.tee`. Essentially, the original iterator should
+    no longer be used after it has been tee'd; only use the tee'd copy.)
 
     If ``k == n`` (tail begins right after the extracted items), the tail
-    is formed from the original iterator at the end of the extraction.
+    is the original iterator at the end of the extraction.
 
     If ``k > n`` (skip some items after the first n), then after extraction,
     the tail is formed by fast-forwarding the iterator using ``drop``.
@@ -303,28 +322,29 @@ def unpack(iterable, n, k=None, fillvalue=None):
     if k < 0:
         raise ValueError("expected k >= 0, got {}".format(k))
     out = []
-    rest = None
+    tl = None
     it = iter(iterable)
     for j in range(n):
         try:
             if j == k:  # tail is desired to overlap with the extracted items
-                it, rest = tee(it)
+                it, tl = tee(it)
             out.append(next(it))
         except StopIteration:  # fewer than n items
             out += [fillvalue] * (n - len(out))
-            def empty():
+            def empty_sequence():
                 yield from ()
-            rest = empty()
-    if not rest:  # avoid replacing empty()
+            tl = empty_sequence()
+            break
+    if not tl:  # avoid replacing empty_sequence()
         if k == n:
-            rest = _makegen(it)
+            tl = it
         elif k > n:
-            rest = drop(k - n, it)
-    out.append(rest)
+            tl = drop(k - n, it)
+    out.append(tl)
     return tuple(out)
 
 def tail(iterable):
-    """Return the tail of an iterable, as a generator.
+    """Return an iterator pointing to the tail of iterable.
 
     Same as ```drop(1, iterable)```.
     """
@@ -345,10 +365,7 @@ def nth(n, iterable, *, default=None):
     """
     if n < 0:
         raise ValueError("expected n >= 0, got {}".format(n))
-    if n == 0:
-        it = iter(iterable)
-    else:
-        it = drop(n - 1, iterable)
+    it = drop(n - 1, iterable) if n else iter(iterable)
     try:
         return next(it)
     except StopIteration:
@@ -366,17 +383,12 @@ def last(iterable, default=None):
     """
     d = deque(iterable, maxlen=1)  # C speed
     return d.pop() if d else default
-#    # slow (Python speed), otherwise fine
-#    item = default
-#    for item in iterable:
-#        pass
-#    return item
 
 def flatten(iterable, pred=None):
     """Recursively remove nested structure from iterable.
 
     Process tuples and lists inside the iterable; pass everything else through
-    (including any generators stored in the iterable).
+    (including any iterators stored in the iterable).
 
     Returns a generator that yields the flattened output.
 
@@ -394,10 +406,13 @@ def flatten(iterable, pred=None):
 
 def flatten1(iterable, pred=None):
     """Like flatten, but process outermost level only."""
-    return _flatten(iterable, pred, recursive=False)
+    if pred:
+        return _flatten(iterable, pred, recursive=False)
+    else:
+        return chain.from_iterable(iterable)  # itertools recipes: fast, no pred
 
 def _flatten(iterable, pred=None, recursive=True):
-    pred = pred or (lambda x: True)
+    pred = pred or (lambda x: True)  # unpythonic.fun.const(True), but dependency loop
     it = iter(iterable)
     for e in it:
         if isinstance(e, (list, tuple)) and pred(e):
@@ -430,16 +445,12 @@ def flatten_in(iterable, pred=None):
                     yield f
             else:
                 t = type(e)
-                new_e = t(flatten_in(e, pred))
-                yield new_e
+                yield t(flatten_in(e, pred))
         else:
             yield e
 
 def iterate1(f, x):
     """Return an infinite generator yielding x, f(x), f(f(x)), ..."""
-#    # elegant FP def, but prone to eventual stack overflow in Python
-#    yield x
-#    yield from iterate1(f, f(x))
     while True:
         yield x
         x = f(x)
@@ -458,10 +469,9 @@ def iterate(f, *args):
         args = f(*args)
 
 def test():
-    from operator import add, mul, itemgetter
-    from functools import partial
+    from operator import add, mul
     from unpythonic.fun import curry, composer, composerc, composel, to1st, rotate, identity
-    from unpythonic.llist import cons, nil, ll
+    from unpythonic.llist import cons, nil, ll, lreverse
 
     # scan/accumulate: lazy fold that yields intermediate results.
     assert tuple(scanl(add, 0, range(1, 5))) == (0, 1, 3, 6, 10)
@@ -484,6 +494,19 @@ def test():
     assert tuple(scanl1(add, (1, 2, 3))) == (1, 3, 6)
     assert tuple(scanr1(add, (1, 2, 3))) == (3, 5, 6)
 
+    from operator import add, mul
+    psums = composer(tail, curry(scanl, add, 0))  # tail to drop the init value
+    pprods = composer(tail, curry(scanl, mul, 1))
+    data = range(1, 5)
+    assert tuple(psums(data)) == (1, 3, 6, 10)
+    assert tuple(pprods(data)) == (1, 2, 6, 24)
+
+    psums = curry(scanl1, add)  # or use the fact the 1-input variant needs no init
+    pprods = curry(scanl1, mul)
+    data = range(1, 5)
+    assert tuple(psums(data)) == (1, 3, 6, 10)
+    assert tuple(pprods(data)) == (1, 2, 6, 24)
+
     def foo(a, b, acc):
         return acc + ((a, b),)
     assert foldl(foo, (), (1, 2, 3), (4, 5)) == ((1, 4), (2, 5))
@@ -544,6 +567,25 @@ def test():
     myadd = lambda x, y: x + y  # can't inspect signature of builtin add
     assert curry(mymap, myadd, ll(1, 2, 3), ll(2, 4, 6)) == ll(3, 6, 9)
 
+    # map_longest. foldr would walk the sequences from the right; use foldl.
+    mymap_longestrev = lambda f: curry(foldl, composerc(cons, f), nil, longest=True)
+    mymap_longest = composerc(lreverse, mymap_longestrev)
+    def noneadd(a, b):
+        if all(x is not None for x in (a, b)):
+            return a + b
+    assert curry(mymap_longest, noneadd, ll(1, 2, 3), ll(2, 4)) == ll(3, 6, None)
+
+    # Adding the missing batteries to the algebra of map and zip.
+    # Note Python's (and Racket's) map is like Haskell's zipWith, but for n inputs.
+    assert tuple(map(add, (1, 2), (3, 4))) == (4, 6)  # builtin
+    assert tuple(mapr(add, (1, 2), (3, 4))) == (6, 4)
+    assert tuple(zip((1, 2, 3), (4, 5, 6), (7, 8))) == ((1, 4, 7), (2, 5, 8))  # builtin
+    assert tuple(zipr((1, 2, 3), (4, 5, 6), (7, 8))) == ((3, 6, 8), (2, 5, 7))
+    assert tuple(map_longest(noneadd, (1, 2, 3), (2, 4))) == (3, 6, None)
+    assert tuple(mapr_longest(noneadd, (1, 2, 3), (2, 4))) == (7, 4, None)
+    assert tuple(zip_longest((1, 2, 3), (2, 4))) == ((1, 2), (2, 4), (3, None))  # itertools
+    assert tuple(zipr_longest((1, 2, 3), (2, 4))) == ((3, 4), (2, 2), (1, None))
+
     reverse_one = curry(foldl, cons, nil)
     assert reverse_one(ll(1, 2, 3)) == ll(3, 2, 1)
 
@@ -561,6 +603,10 @@ def test():
     assert mysum(append_two(a, b)) == 10
     assert myprod(b) == 12
 
+    packtwo = lambda a, b: ll(a, b)  # using a tuple return value here would confuse curry.
+    assert foldl(composerc(cons, packtwo), nil, (1, 2, 3), (4, 5), longest=True) == \
+           ll(ll(3, None), ll(2, 5), ll(1, 4))
+
     def msqrt(x):  # multivalued sqrt
         if x == 0.:
             return (0.,)
@@ -602,8 +648,6 @@ def test():
     assert a == tuple(range(3))
     assert b == ()
 
-    assert tuple(zipr((1, 2, 3), (4, 5, 6), (7, 8))) == ((3, 6, 8), (2, 5, 7))
-
     @rotate(1)
     def zipper(acc, *rest):   # so that we can use the *args syntax to declare this
         return acc + (rest,)  # even though the input is (e1, ..., en, acc).
@@ -642,13 +686,28 @@ def test():
     assert tuple(flatten(data, is_nested))    == (((1, 2), ((3, 4), (5, 6)), 7), (8, 9), (10, 11))
     assert tuple(flatten_in(data, is_nested)) == (((1, 2), (3, 4), (5, 6), 7),   (8, 9), (10, 11))
 
+    # lazy unpack from a sequence
+    a, b, c, tl = unpack(3, range(5))
+    assert a == 0 and b == 1 and c == 2
+    assert next(tl) == 3
+
+    # lazy unpack falling off the end of a sequence
+    a, b, c, tl = unpack(3, range(2))
+    assert a == 0 and b == 1 and c is None
+    try:
+        next(tl)
+    except StopIteration:
+        pass
+    else:
+        assert False  # the tail should be empty
+
     # unpacking of generators - careful!
     def mygen():
         for x in range(10, 16):
             yield x
     def dostuff(s):
-        a, b, t = unpack(s, 2, 0)  # peek two, set t to 0th tail (the original s)
-        return t
+        a, b, tl = unpack(2, s, k=0)  # peek two, set tl to 0th tail (the original s)
+        return tl
     g = mygen()
     dostuff(g)  # gotcha: advances g!
     assert next(g) == 12
@@ -677,7 +736,7 @@ def test():
     assert tuple(collatz(10)) == (10, 5, 16, 8, 4, 2, 1)
     assert tuple(collatz(30)) == (30, 15, 46, 23, 70, 35, 106, 53, 160, 80, 40, 20, 10, 5, 16, 8, 4, 2, 1)
     def len_gt(k, s):
-        a, _ = unpack(drop(k, s), 1)
+        a, _ = unpack(1, drop(k, s))
         return a  # None if no item
     islong = curry(len_gt, 15)
     assert sum(1 for n in range(1, 101) if islong(collatz(n))) == 66
@@ -741,7 +800,7 @@ def test():
     def within(eps, s):
         while True:
             # unpack with peek (but be careful, the rewinded tail is a tee'd copy)
-            a, b, s = unpack(s, 2, 1)
+            a, b, s = unpack(2, s, k=1)
             if abs(a - b) < eps:
                 return b
     def differentiate_with_tol(h0, f, x, eps):
@@ -750,7 +809,7 @@ def test():
 
     def order(s):
         """Estimate asymptotic order of s, consuming the first three terms."""
-        a, b, c, _ = unpack(s, 3)
+        a, b, c, _ = unpack(3, s)
         return round(log2(abs((a - c) / (b - c)) - 1))
     def eliminate_error(n, s):
         """Eliminate error term of given asymptotic order n.
@@ -758,7 +817,7 @@ def test():
         The stream s must be based on halving h at each step
         for the formula used here to work."""
         while True:
-            a, b, s = unpack(s, 2, 1)
+            a, b, s = unpack(2, s, k=1)
             yield (b*2**n - a) / (2**(n - 1))
     def improve(s):
         """Eliminate asymptotically dominant error term from s.
@@ -796,7 +855,7 @@ def test():
     # https://en.wikipedia.org/wiki/Series_acceleration#Euler%27s_transform
     def euler_transform(s):
         while True:
-            a, b, c, s = unpack(s, 3, 1)
+            a, b, c, s = unpack(3, s, k=1)
             yield c - ((c - b)**2 / (a - 2*b + c))
     faster_pi_stream = euler_transform(pi_stream)
 
